import Foundation

/// Instances of this class are capable of encoding types into CSV files.
@dynamicMemberLookup public class CSVEncoder {
    /// Wrap all configurations in a single easy-to-use structure.
    private final var configuration: Configuration
    /// A dictionary you use to customize the encoding process by providing contextual information.
    open var userInfo: [CodingUserInfoKey:Any]
    
    /// Creates a CSV encoder with the default configuration values.
    /// - parameter configuration: Optional configuration values for the encoding process.
    public init(configuration: Configuration = .init()) {
        self.configuration = configuration
        self.userInfo = .init()
    }
    
    /// Creates a CSV encoder and passes the default configuration values to the closure.
    /// - parameter configuration: A closure receiving the default configuration values and returning (by `inout`) a tweaked version of them.
    @inlinable public convenience init(configuration: (inout Configuration)->Void) {
        var config = Configuration()
        configuration(&config)
        self.init(configuration: config)
    }
    
    /// Gives direct access to all the encoder's configuration values.
    /// - parameter member: Writable key path for the decoder's configuration value.
    public final subscript<V>(dynamicMember member: WritableKeyPath<CSVEncoder.Configuration,V>) -> V {
        get { self.configuration[keyPath: member] }
        set { self.configuration[keyPath: member] = newValue }
    }
}

extension CSVEncoder {
    /// Returns a CSV-encoded representation of the value you supply.
    /// - parameter value: The value to encode as CSV.
    /// - returns: `Data` blob with the CSV representation of `value`.
    open func encode<T:Encodable>(_ value: T) throws -> Data {
        let writer = try CSVWriter(configuration: self.configuration.writerConfiguration)
        let sink = try ShadowEncoder.Sink(writer: writer, configuration: self.configuration, userInfo: self.userInfo)
        try value.encode(to: ShadowEncoder(sink: sink, codingPath: []))
        try sink.completeEncoding()
        return try writer.data()
    }
    
    /// Returns a CSV-encoded representation of the value you supply.
    /// - parameter value: The value to encode as CSV.
    /// - returns: `String` with the CSV representation of `value`.
    open func encode<T:Encodable>(_ value: T, into: String.Type) throws -> String {
        let data = try self.encode(value)
        let encoding = self.configuration.writerConfiguration.encoding ?? .utf8
        return String(data: data, encoding: encoding)!
    }
    
    /// Returns a CSV-encoded representation of the value you supply.
    /// - parameter value: The value to encode as CSV.
    /// - parameter fileURL: The file receiving the encoded values.
    /// - parameter append: In case an existing file is under the given URL, this Boolean indicates that the information will be appended to the file (`true`), or the file will be overwritten (`false`).
    open func encode<T:Encodable>(_ value: T, into fileURL: URL, append: Bool = false) throws {
        let writer = try CSVWriter(fileURL: fileURL, append: append, configuration: self.configuration.writerConfiguration)
        let sink = try ShadowEncoder.Sink(writer: writer, configuration: self.configuration, userInfo: self.userInfo)
        try value.encode(to: ShadowEncoder(sink: sink, codingPath: []))
        try sink.completeEncoding()
    }
}

extension CSVEncoder: Failable {
    /// The type of error raised by the CSV writer.
    public enum Error: Int {
        /// Some of the configuration values provided are invalid.
        case invalidConfiguration = 1
        /// The encoding coding path is invalid.
        case invalidPath = 2
        /// An error occurred on the encoder buffer.
        case bufferFailure = 4
    }
    
    public static var errorDomain: String {
        "Writer"
    }
    
    public static func errorDescription(for failure: Error) -> String {
        switch failure {
        case .invalidConfiguration: return "Invalid configuration"
        case .invalidPath: return "Invalid coding path"
        case .bufferFailure: return "Invalid buffer state"
        }
    }
}

